import sys
if sys.version_info >= (3, 11):
    from tomllib import load # 需要 Python3.11 及其以上版本
else:
    from tomlkit import load # 需要安装 tomlkit
from dataclasses import dataclass
from pathlib import Path
# from taolib.utils.tree import Bunch

@dataclass
class Config:
    root_dir: Path | str # 文档的根目录
    name: str = "_config.toml" # 文档目录名称

    def __post_init__(self):
        self.root_dir = Path(self.root_dir)
        # 加载文档的配置信息
        with open(self.root_dir/self.name, "rb") as f:
            self.options = load(f)
    # def write(self, name: str):
    #     return 


try:
    from .qt5 import QtCore, QtGui, QtWidgets, Signal, Slot, QtPrintSupport, QtQml
except:
    from .qt6 import QtCore, QtGui, QtWidgets, Signal, Slot, QtPrintSupport, QtQml


def run(window_type, *args, **kwargs):
    '''运行 GUI'''
    #import sys
    # 实例化一个QApplication对象，用于GUI的初始化
    app = QtWidgets.QApplication([]) # QtWidgets.QApplication(sys.argv)
    # 实例化一个 Window 对象，用于定义一个图形窗口
    window = window_type(*args, **kwargs)
    # # 适配 Retina 显示屏（选写）.
    # app.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)
    # app.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
    window.show() # 显示主窗口
    app.exec_()  # 当GUI产生退出信号时Python程序结束
    # 利用内置模块sys的exit()方法侦听GUI的退出信号，以便关闭Python程序。
    #sys.exit(app.exec_())  
    


from tkinter import Tk, Toplevel, ttk


class ToolTip:
    def __init__(self, widget):
        self.widget = widget
        self.tip_window = None

    def show_tip(self, tip_text):
        "Display text in a tooltip window"
        if self.tip_window or not tip_text:
            return
        x, y, _cx, cy = self.widget.bbox("insert")      # get size of widget
        x = x + self.widget.winfo_rootx() + 25          # calculate to display tooltip
        y = y + cy + self.widget.winfo_rooty() + 25     # below and to the right
        # create new tooltip window
        self.tip_window = tw = Toplevel(self.widget)
        # remove all Window Manager (wm) decorations
        tw.wm_overrideredirect(True)
        # tw.wm_overrideredirect(False)                 # uncomment to see the effect
        tw.wm_geometry("+%d+%d" % (x, y))               # create window size

        label = ttk.Label(tw, text=tip_text, justify='left',
                          background="#ffffe0", relief='solid',
                          font=("tahoma", "8", "normal"))
        label.pack(ipadx=1)

    def hide_tip(self):
        tw = self.tip_window
        self.tip_window = None
        if tw:
            tw.destroy()


def create_ToolTip(widget, text):
    toolTip = ToolTip(widget)       # create instance of class

    def enter(event):
        toolTip.show_tip(text)

    def leave(event):
        toolTip.hide_tip()
    widget.bind('<Enter>', enter)   # bind mouse events
    widget.bind('<Leave>', leave)



"""下载一些东西"""
from github import Github
from io import BytesIO


def get_github_content(user_name, repo_name, filename):
    """"""
    g = Github()
    user = g.get_user(user_name)
    repo = user.get_repo(repo_name)
    contents = repo.get_contents(filename)
    return contents.decoded_content.decode()


def iter_github_bytes(user_name, repo_name, root_dir):
    """"""
    g = Github()
    user = g.get_user(user_name)
    repo = user.get_repo(repo_name)
    contents = repo.get_contents(root_dir)
    for content in contents:
        yield BytesIO(content.decoded_content)



from tkinter import Tk, Menu, filedialog
from PIL import Image, ImageTk


class WindowMeta(Tk):
    def __init__(self):
        super().__init__()
        self._menu_init()
        self.create_menu()

    def _menu_init(self):
        # The settings menu cannot pop up from the window.
        self.option_add('*tearOff', False)

    def create_menu(self):
        self.menu_bar = Menu(self)
        self['menu'] = self.menu_bar  # Or `root.config(menu=menubar)`
        self._create_menu()

    def _create_menu(self):
        self._create_file_menu()
        self._create_edit_menu()

    def _create_edit_menu(self):
        self.edit_bar = Menu(self.menu_bar)
        self.move_menu = Menu(self.edit_bar)
        self.delete_menu = Menu(self.edit_bar)
        self.menu_bar.add_cascade(label="Edit", menu=self.edit_bar)
        self.edit_bar.add_cascade(label="move", menu=self.move_menu)
        self.edit_bar.add_cascade(label="delete", menu=self.delete_menu)

    def _create_file_menu(self):
        self.file_bar = Menu(self.menu_bar)
        self.image_menu = Menu(self.file_bar)
        self.tags_menu = Menu(self.file_bar)
        self.menu_bar.add_cascade(label="File", menu=self.file_bar)

    def seek_folder_name(self):
        return filedialog.askdirectory()

    def seek_filename(self):
        return filedialog.askopenfilename()

    def seek_filenames(self):
        return filedialog.askopenfilenames()

    def name2image(self, image_name):
        I = Image.open(image_name)
        _ = ImageTk.PhotoImage(I)
        return ImageTk.PhotoImage(I)


class Window(WindowMeta):
    def __init__(self):
        super().__init__()
        self.file_name = None
        self.I = None

    def update_edit_menu(self, graph):
        kw_menu = {
            'variable': graph.edit_var,
            'command': graph.bind_graph
        }
        self.edit_bar.add_radiobutton(label=f"drawing", **kw_menu)
        [self.move_menu.add_radiobutton(
            label=f"move/{option}", **kw_menu) for option in graph.tags_dict]
        [self.delete_menu.add_radiobutton(
            label=f"delete/{option}", **kw_menu) for option in graph.tags_dict]

    def update_file(self):
        self.file_name = self.seek_filename()

    def update_image(self):
        self.I = self.name2image(self.file_name)

    def reload_image(self, graph):
        graph.create_image(0, 0, anchor='nw', image=self.I, tags='image')

    def load_image(self, graph):
        self.update_file()
        self.update_image()
        self.reload_image(graph)

    def update_file_menu(self, graph):
        self.file_bar.add_radiobutton(
            label="Seek filename", command=lambda : self.load_image(graph))




from .. import QtWidgets, QtGui, QtCore


Qt = QtCore.Qt
QGraphicsItem = QtWidgets.QGraphicsItem
QColor = QtGui.QColor
QRectF = QtCore.QRectF


class RectHandle(QtWidgets.QGraphicsRectItem):
    # handles 按照顺时针排列
    handle_names = ('left_top', 'middle_top', 'right_top', 'right_middle',
                    'right_bottom', 'middle_bottom', 'left_bottom', 'left_middle')
    # 设定在控制点上的光标形状
    handle_cursors = {
        0: Qt.SizeFDiagCursor,
        1: Qt.SizeVerCursor,
        2: Qt.SizeBDiagCursor,
        3: Qt.SizeHorCursor,
        4: Qt.SizeFDiagCursor,
        5: Qt.SizeVerCursor,
        6: Qt.SizeBDiagCursor,
        7: Qt.SizeHorCursor
    }
    offset = 6.0  # 外边界框相对于内边界框的偏移量，也是控制点的大小
    #min_size = 8 * offset  # 矩形框的最小尺寸

    def update_handles_pos(self):
        """
        更新控制点的位置
        """
        o = self.offset  # 偏置量
        s = o*2  # handle 的大小
        b = self.rect()  # 获取内边框
        x1, y1 = b.left(), b.top()  # 左上角坐标
        offset_x = b.width()/2
        offset_y = b.height()/2
        # 设置 handles 的位置
        self.handles[0] = QRectF(x1-o, y1-o, s, s)
        self.handles[1] = self.handles[0].adjusted(offset_x, 0, offset_x, 0)
        self.handles[2] = self.handles[1].adjusted(offset_x, 0, offset_x, 0)
        self.handles[3] = self.handles[2].adjusted(0, offset_y, 0, offset_y)
        self.handles[4] = self.handles[3].adjusted(0, offset_y, 0, offset_y)
        self.handles[5] = self.handles[4].adjusted(-offset_x, 0, -offset_x, 0)
        self.handles[6] = self.handles[5].adjusted(-offset_x, 0, -offset_x, 0)
        self.handles[7] = self.handles[6].adjusted(0, -offset_y, 0, -offset_y)


class RectItem(RectHandle):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.handles = {}  # 控制点的字典
        self.setAcceptHoverEvents(True)  # 设定为接受 hover 事件
        self.setFlags(QGraphicsItem.ItemIsSelectable |  # 设定矩形框为可选择的
                      QGraphicsItem.ItemSendsGeometryChanges |  # 追踪图元改变的信息
                      QGraphicsItem.ItemIsFocusable |  # 可聚焦
                      QGraphicsItem.ItemIsMovable)  # 可移动
        self.update_handles_pos()  # 初始化控制点
        self.reset_Ui()  # 初始化 UI 变量

    def reset_Ui(self):
        '''初始化 UI 变量'''
        self.handleSelected = None  # 被选中的控制点
        self.mousePressPos = None  # 鼠标按下的位置
        #self.mousePressRect = None  # 鼠标按下的位置所在的图元的外边界框

    def boundingRect(self):
        """
        限制图元的可视化区域，且防止出现图元移动留下残影的情况
        """
        o = self.offset
        # 添加一个间隔为 o 的外边框
        return self.rect().adjusted(-o, -o, o, o)

    def paint(self, painter, option, widget=None):
        """
        Paint the node in the graphic view.
        """
        painter.setBrush(QtGui.QBrush(QColor(255, 0, 0, 100)))
        painter.setPen(QtGui.QPen(QColor(0, 0, 0), 0, Qt.SolidLine))
        painter.drawRect(self.rect())

        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        painter.setBrush(QtGui.QBrush(QColor(255, 255, 0, 200)))
        painter.setPen(QtGui.QPen(QColor(0, 0, 0, 255), 0,
                                  Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin))

        for shape in self.handles.values():
            if self.isSelected():
                painter.drawEllipse(shape)

    def handle_at(self, point):
        """
        返回给定 point 下的控制点 handle
        """
        for k, v, in self.handles.items():
            if v.contains(point):
                return k
        return

    def hoverMoveEvent(self, event):
        """
        当鼠标移到该 item（未按下）上时执行。
        """
        super().hoverMoveEvent(event)
        handle = self.handle_at(event.pos())
        cursor = self.handle_cursors[handle] if handle in self.handles else Qt.ArrowCursor
        self.setCursor(cursor)

    def hoverLeaveEvent(self, event):
        """
        当鼠标离开该形状（未按下）上时执行。
        """
        super().hoverLeaveEvent(event)
        self.setCursor(Qt.ArrowCursor)  # 设定鼠标光标形状

    def mousePressEvent(self, event):
        """
        当在 item 上按下鼠标时执行。
        """
        super().mousePressEvent(event)
        self.handleSelected = self.handle_at(event.pos())
        if self.handleSelected in self.handles:
            self.mousePressPos = event.pos()

    def mouseReleaseEvent(self, event):
        """
        Executed when the mouse is released from the item.
        """
        super().mouseReleaseEvent(event)
        self.update()
        self.reset_Ui()

    def mouseMoveEvent(self, event):
        """
        Executed when the mouse is being moved over the item while being pressed.
        """
        if self.handleSelected in self.handles:
            self.interactiveResize(event.pos())
        else:
            super().mouseMoveEvent(event)

    def interactiveResize(self, mousePos):
        """
        Perform shape interactive resize.
        """
        rect = self.rect()
        self.prepareGeometryChange()
        # movePos = mousePos - self.mousePressPos
        # move_x, move_y = movePos.x(), movePos.y()
        if self.handleSelected == 0:
            rect.setTopLeft(mousePos)
        elif self.handleSelected == 1:
            rect.setTop(mousePos.y())
        elif self.handleSelected == 2:
            rect.setTopRight(mousePos)
        elif self.handleSelected == 3:
            rect.setRight(mousePos.x())
        elif self.handleSelected == 4:
            rect.setBottomRight(mousePos)
        elif self.handleSelected == 5:
            rect.setBottom(mousePos.y())
        elif self.handleSelected == 6:
            rect.setBottomLeft(mousePos)
        elif self.handleSelected == 7:
            rect.setLeft(mousePos.x())
        self.setRect(rect)
        self.update_handles_pos()



try:
    # PySide2
    from PySide2 import QtGui, QtWidgets, QtCore, QtPrintSupport, QtQml
    from PySide2.QtCore import Signal, Slot
except ImportError:
    # PyQt5
    from PyQt5 import QtGui, QtWidgets, QtCore, QtPrintSupport, QtQml
    from PyQt5.QtCore import pyqtSignal as Signal, pyqtSlot as Slot


from abc import ABC, abstractmethod


class Validator(ABC):
    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        NotImplemented

class OneOf(Validator):
    def __init__(self, *options):
        self.options = set(options)

    def validate(self, value):
        if value not in self.options:
            raise ValueError(f'Expected {value!r} to be one of {self.options!r}')

class Number(Validator):
    def __init__(self, minvalue=None, maxvalue=None):
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def validate(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError(f'Expected {value!r} to be an int or float')
        if self.minvalue is not None and value < self.minvalue:
            raise ValueError(
                f'Expected {value!r} to be at least {self.minvalue!r}'
            )
        if self.maxvalue is not None and value > self.maxvalue:
            raise ValueError(
                f'Expected {value!r} to be no more than {self.maxvalue!r}'
            )

class String(Validator):
    def __init__(self, minsize=None, maxsize=None, predicate=None):
        self.minsize = minsize
        self.maxsize = maxsize
        self.predicate = predicate

    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError(f'Expected {value!r} to be an str')
        if self.minsize is not None and len(value) < self.minsize:
            raise ValueError(
                f'Expected {value!r} to be no smaller than {self.minsize!r}'
            )
        if self.maxsize is not None and len(value) > self.maxsize:
            raise ValueError(
                f'Expected {value!r} to be no bigger than {self.maxsize!r}'
            )
        if self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate} to be true for {value!r}'
            )


if __name__ == "__main__":
    class Component:
        name = String(minsize=3, maxsize=10, predicate=str.isupper)
        kind = OneOf('wood', 'metal', 'plastic')
        quantity = Number(minvalue=0)

        def __init__(self, name, kind, quantity):
            self.name = name
            self.kind = kind
            self.quantity = quantity


from dataclasses import dataclass
import numpy as np

# np.int15 = np.dtype([("int15", "b", 15)]) # 等价于 np.int15 = np.dtype([("int15", "i1", 15)])
np.int8e = np.dtype([("int8e", np.int8, 16)]) # 针对深度可分离卷积的扩展权重

@dataclass(eq=True, order=True)
class TensorType:
    shape: tuple[int]
    dtype: str
    name: str = "data"

    def __post_init__(self):
        if self.dtype == "int8e":
            self.dtype = np.int8e

    @property
    def empty(self):
        '''空的输出张量'''
        return np.empty(shape=self.shape,
                        dtype=self.dtype)

    @property
    def nbytes(self):
        '''输出张量的存储字节数'''
        return self.empty.nbytes

    @property
    def nelement(self):
        '''输出张量的元素个数'''
        return self.empty.size

    def weak_eq(self, other):
        """不考虑对名称的依赖"""
        assert_shape = self.shape == other.shape
        assert_dtype = self.dtype == other.dtype
        return assert_shape and assert_dtype



from pathlib import Path
import logging

logger = logging.getLogger(__name__)


def mkdir(root: str|Path, mode=0o755, parents=True, exist_ok=True):
    """创建文件夹"""
    root = Path(root)
    if not root.exists():
        root.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        logger.info(f"{root} 创建成功")
    else:
        logger.info(f"{root} 已存在无需创建")



from invoke import Collection

from . import doc
from . import release

# def site(name='docs', target='output/html'):
#     if name == 'docs':
#         out = ''
#     else:
#         out = name
#     _config = {"sphinx": {
#         "source": name,
#         "target": f"{target}/{out}"
#     }}
#     namespace = Collection()
#     namespace.add_collection(release)
#     namespace.add_collection(docs)
#     namespace.collections['docs'].configure(_config)
#     return namespace

def site(source='docs', target='output/html', children=''):
    source = children if children else source
    _config = {"sphinx": {
        "source": source,
        "target": f"{target}/{children}"
    }}
    namespace = Collection()
    namespace.add_collection(release)
    namespace.add_collection(doc)
    namespace.collections['doc'].configure(_config)
    return namespace


try:
    # PySide6
    from PySide6 import QtGui, QtWidgets, QtCore, QtPrintSupport, QtQml
    from PySide6.QtCore import Signal, Slot
except ImportError:
    # PyQt6
    from PyQt6 import QtGui, QtWidgets, QtCore, QtPrintSupport, QtQml
    from PyQt6.QtCore import pyqtSignal as Signal, pyqtSlot as Slot


from dash import html

__all__ = ['create_nav']


def create_icon_button(icon_class, icon_name, button_id=''):
    icon = html.I(f' {icon_name}', className=icon_class)
    button = html.Button(icon,
                         className='w3-button w3-padding-small',
                         type='button',
                         id=button_id)
    return button


def create_icon_label(icon_class, icon_name, htmlFor):
    icon = html.I(f' {icon_name}', className=icon_class)
    label = html.Label(icon,
                       className='w3-bar-item w3-button',
                       id=htmlFor,
                       htmlFor=htmlFor)
    return label


def create_label_li(label):
    li = html.Li(label)
    return li


def create_ul(li_list):
    ul = html.Ul(li_list,
                 className="w3-dropdown-content w3-bar-block w3-card-4")
    return ul


def create_menu(labels, button_icon_class, button_icon_name):
    li_list = [create_label_li(label) for label in labels]
    ul = create_ul(li_list)
    button = create_icon_button(button_icon_class, button_icon_name)
    return html.Section([button, ul], className='w3-dropdown-hover')


def create_nav():
    file_labels = [create_icon_label(*k) for k in [['fas fa-file', 'open file', 'open_file'],
                                                   ['fas fa-trash',
                                                       'Trash', 'trash'],
                                                   ['fas fa-window-close', 'Close', 'close']]]
    file_icon_name = 'file'
    file_icon_class = 'far fa-file'
    file_sec = create_menu(file_labels, file_icon_class, file_icon_name)

    edit_labels = [create_icon_label(*k) for k in [['fas fa-film', 'cv', 'cv'],
                                                   ['fas fa-map-marked-alt',
                                                       'radar', 'radar'],
                                                   ['fas fa-car', 'vehicle', 'vehicle']]]
    edit_icon_name = 'edit'
    edit_icon_class = 'fas fa-edit'
    edit_sec = create_menu(edit_labels, edit_icon_class, edit_icon_name)

    display_labels = [create_icon_label(*k) for k in [['fas fa-film', 'display cv', 'display_cv'],
                                                      ['fas fa-map-marked-alt',
                                                       'display radar',
                                                       'display_radar'],
                                                      ['fas fa-car', 'display vehicle', 'display_vehicle']]]
    display_icon_name = 'display'
    display_icon_class = 'fas fa-eye'
    display_sec = create_menu(
        display_labels, display_icon_class, display_icon_name)

    run_labels = [create_icon_label(*k) for k in [['fas fa-play', 'start', 'start'],
                                                  ['fas fa-stop', 'stop', 'stop']]]
    run_icon_name = 'run'
    run_icon_class = 'far fa-play-circle'
    run_sec = create_menu(run_labels, run_icon_class, run_icon_name)

    nav = html.Nav([file_sec, edit_sec, display_sec, run_sec],
                   className='w3-bar w3-bottombar w3-border-top')
    return nav



"""处理 bunch 数据的接口"""

class Bunch(dict):
    """用于创建类字典的可以使用点号访问属性的对象，并且可以处理嵌套的字典"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self # 这意味着 Bunch 类的实例将具有与字典相同的行为，可以使用点符号访问和修改其键值对
        self._convert_nested_dicts()

    def _convert_nested_dicts(self):
        """合并嵌套的字典"""
        for k, v in self.__dict__.items():
            if isinstance(v, dict):
                self.__dict__[k] = Bunch(**v)  # 将字典转换为 Bunch 对象
            elif isinstance(v, Bunch):
                v._convert_nested_dicts()  # 递归处理嵌套的 Bunch 对象
                     
    def merge(self, other):
        """提供递归合并功能"""
        other = Bunch(**other)
        for k, v in other.items():
            if k not in self:
                self[k] = other[k]
            else:
                if not isinstance(self[k], dict) and not isinstance(v, dict):
                    self[k] = v
                elif isinstance(self[k], dict) and isinstance(v, dict):
                    self[k].update(v)
                else:
                    raise TypeError(f"{other}不支持合并")



from pathlib import Path

def rename_suffix(root, old, new):
    '''批量修改文件后缀'''
    root = Path(root)
    for path in root.rglob(f'*{old}'):
        new_name = path.name.split(old)[0]+new
        path.rename(path.parent/new_name)


"""任务型事务辅助函数"""
import asyncio
# import logging
import textwrap

async def communicate_cmd(cmd: str, cwd: str|None=None, indent: int=2,  **kwds)->str:
    """运行命令行执行任务

    Args:
        cmd: 命令行任务
        cwd: 当前工作目录
        indent: 命令行打印信息缩进位数

    Return:
        返回命令任务执行的打印信息
    """
    _ = lambda text: textwrap.indent(text, prefix=" "*indent) # 美化打印信息
    proc = await asyncio.create_subprocess_shell(
        cmd, 
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=cwd, **kwds)
    stdout, stderr = await proc.communicate(input=None)
    std_data = f'[{cmd!r} 退出状态码 {proc.returncode}]\n'
    if stdout:
        std_data += f'[stdout]:\n{_(stdout.decode())}\n' 
    if stderr:
        std_data += f'[stdinfo]:\n{_(stderr.decode())}\n'
    return std_data

# async def log_cmd(cmd: str, cwd: str|None=None, indent: int=2, logger_name="run", **kwds)->str:
#     """运行命令行执行任务，并记录打印信息

#     Args:
#         cmd: 命令行任务
#         cwd: 当前工作目录
#         indent: 命令行打印信息缩进位数

#     Return:
#         返回命令任务执行的打印信息
#     """
#     logger = logger = logging.getLogger(logger_name)
#     hd = logging.StreamHandler()
#     hd.setFormatter(logging.Formatter("%(message)s"))
#     logger.addHandler(hd)
#     proc = await asyncio.create_subprocess_shell(
#         cmd, 
#         stdin=asyncio.subprocess.PIPE,
#         stdout=asyncio.subprocess.PIPE,
#         stderr=asyncio.subprocess.PIPE,
#         cwd=cwd, **kwds)
#     # 获取实时输出/错误信息
#     if not proc.stdout.at_eof():
#         for line in iter(proc.stdout.readline, b''):
#             line = await line
#             logger.info(line.decode())
#             if proc.stdout.at_eof():
#                 break
#     if not proc.stderr.at_eof():
#         for line in iter(proc.stderr.readline, b''):
#             line = await line
#             logger.info(line.decode())
#             if proc.stderr.at_eof():
#                 break
#     # 等待命令执行完成
#     proc.wait()


from pathlib import Path
import sys
import logging
from shutil import rmtree
from tempfile import mkdtemp
from invoke import task
from invoke.context import Context

logger = logging.getLogger(__name__)

pty = sys.platform != 'win32'

@task
def clean(ctx: Context):
    """清除文档构建目标目录，以便下次构建是干净的。"""
    output = Path(ctx.sphinx.target)
    if output.exists():
        logger.info(f'删除 {output}')
        rmtree(output)


@task(default=True)
def build(ctx: Context,
          opts: str = "",
          language: str|None = None,
          source: str|None = None,
          target: str|None = None,
          nitpick: bool = False):
    """构建项目的 Sphinx 文档。
    
    Args:
        opts: Extra sphinx-build options/args
        nitpick: Build with stricter warnings/errors enabled
        source: Source directory; overrides config setting
        target: Output directory; overrides config setting
    """
    source = source or ctx.sphinx.source
    target = target or ctx.sphinx.target
    if language:
        opts = f'-D language={language}'
        target = f'{target}/{language}'
    if nitpick:
        opts += " -n -W -T"
    cmd = f"sphinx-build {opts} {source} {target}"
    logger.info(f"{source} => {target}")
    ctx.run(cmd, pty=pty)

@task
def intl(ctx: Context, language: str = 'en'):
    '''Update the POT file and invoke the `sphinx-intl update` command

    Only used with `invoke intl.update`
    '''
    opts = "-b gettext"
    target = Path(ctx.sphinx.target).parent / 'gettext'
    if language == 'en':
        clean(ctx)
        build(ctx, target=target, opts=opts)
    elif language:
        if not Path(target).exists():
            build(ctx, target=target, opts=opts)
        ctx.run(
            f'sphinx-intl update -p {target} -l {language}'
        )
        # for DIR in ['pages', 'posts', 'shop']:
        #     rmtree(f'locales/{language}/LC_MESSAGES/{DIR}/')

@task
def doctest(ctx: Context):
    """
    Run Sphinx' doctest builder.

    This will act like a test run, displaying test results & exiting nonzero if
    all tests did not pass.

    A temporary directory is used for the build target, as the only output is
    the text file which is automatically printed.
    """
    tmpdir = mkdtemp()
    try:
        opts = "-b doctest"
        target = tmpdir
        clean(ctx)
        build(ctx, target=target, opts=opts)
    finally:
        rmtree(tmpdir)

@task
def tree(ctx: Context):
    """
    Display documentation contents with the 'tree' program.
    """
    ignore = ".git|*.pyc|*.swp|dist|*.egg-info|_static|_build|_templates"
    ctx.run(f'tree -Ca -I "{ignore}" {ctx.sphinx.source}')



from invoke import Collection
from . import doc

def sites(source='doc', target='.temp/html', children=''):
    source = children if children else source
    _config = {"sphinx": {
        "source": source,
        "target": f"{target}/{children}"
    }}
    namespace = Collection()
    namespace.add_collection(doc)
    namespace.collections['doc'].configure(_config)
    return namespace



from pathlib import Path
from zipfile import ZipFile


def decode_path(path: str|Path):
    '''将乱码的路径编码为 UTF8

    Args:
        path: Path 的实例
    '''
    try:
        path_name = path.decode('utf-8')
    except:
        path_name = path.encode('437').decode('gbk')
        path_name = path_name.encode('utf-8').decode('utf-8')
    return path_name


def extract(zip_name: str, out_dir: str|Path):
    '''将 zip_name 文件解压到 out_dir 目录
    '''
    out_dir = Path(out_dir)
    with ZipFile(zip_name, allowZip64=True) as Z:
        # 排除目录文件
        file_iter = (file for file in Z.filelist if not file.is_dir())
        for file in file_iter:
            # 编码文件名称为 utf 格式
            file.filename = decode_path(file.filename)
            Z.extract(file, out_dir/file.filename)


def extract_all(zip_root: str|Path, out_dir: str|Path):
    '''解压 zip_root 目录下的全部 zip 文件到 out_dir 目录'''
    for zip_dir in Path(zip_root).rglob('*.zip'):
        # 解压单个数据集
        extract(zip_dir, out_dir)
    print("全部解压完成！")



import logging
from dash import Dash
import socket

# def interpolate_str(template, **data):
#     s = template
#     for k, v in data.items():
#         key = "{%" + k + "%}"
#         s = s.replace(key, v)
#     return s


def get_Host_name_IP():
    try:
        host_name = socket.gethostname()
        host_ip = socket.gethostbyname(host_name)
        logging.info(f"Hostname: {host_name}\nIP: {host_ip}")
    except:
        logging.info("Unable to get Hostname and IP")
        host_ip = 'localhost'
    return host_ip


# Driver code
host_ip = get_Host_name_IP()  # Function call

# This code is conributed by "Sharad_Bhardwaj".


NAME = __name__

META_TAGS = [
    {
        'name': 'description',
        'content': 'Using AI to develop anything interesting'
    },
    # A tag that tells the browser not to scale
    # desktop widths to fit mobile screens.
    # Sets the width of the viewport (browser)
    # to the width of the device, and the zoom level
    # (initial scale) to 1.
    #
    # Necessary for "true" mobile support.
    {
        'name': 'viewport',
        'content': 'width=device-width, initial-scale=1, shrink-to-fit=no'
    }
]


def create_app(name=NAME, title='Dash', external_stylesheets=None, **kwargs):
    # external_scripts = ['https://www.google-analytics.com/analytics.js']
    external_scripts = [
        "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.5/MathJax.js?config=TeX-MML-AM_CHTML"]

    _external_stylesheets = ['https://xinetzone.github.io/Font-Awesome/css/all.css',
                             'https://xinetzone.github.io/w3css/4/w3.css',
                             'https://xinetzone.github.io/xinet-css/tabs.css']
    if external_stylesheets==None:
        external_stylesheets = _external_stylesheets
    kw = {
        'meta_tags': META_TAGS,
        'external_stylesheets': external_stylesheets,
        'external_scripts': external_scripts
    }
    kwargs.update(kw)
    app = Dash(name, title=title, **kwargs)
    return app


index_string_template = '''<!DOCTYPE html>
<html lang="zh">
    <head>
        {%metas%}
        <title>{%title%}</title>
        {%favicon%}
        {%css%}
    </head>
    <body>
        {%app_entry%}
        <article>
            {%config%}
            {%scripts%}
            {%renderer%}
        </article>
    </body>
</html>
'''


async def run_server(app, layout, host=host_ip, port=8050, mode='external', debug=True, **kw):
    # host='0.0.0.0' 、 127.0.0.1
    # app = create_app()
    app.index_string = index_string_template
    app.layout = layout
    app.run_server(mode, host=host, port=port,
                   debug=debug, use_reloader=False, **kw)



'''临时代码'''
import sys
from pathlib import Path
from shutil import rmtree
from invoke import Collection, task


@task(name='clean')
def _clean(c):
    output = Path(c.sphinx.target)
    if output.exists():
        print(f'delete {output}')
        rmtree(output)


@task(
    default=True,
    help={
        "opts": "Extra sphinx-build options/args",
        "nitpick": "Build with stricter warnings/errors enabled",
        "source": "Source directory; overrides config setting",
        "target": "Output directory; overrides config setting",
    },
)
def build(c,
          opts=None,
          language=None,
          source=None,
          target=None,
          nitpick=False):
    """
    Build the project's Sphinx docs.
    """
    if opts is None:
        opts = ""
    source = source or c.sphinx.source
    target = target or c.sphinx.target
    if language:
        opts = f'-D language={language}'
        target = f'{target}/{language}'
    if nitpick:
        opts += " -n -W -T"
    cmd = f"sphinx-build {opts} {source} {target}"
    c.run(cmd)


@task
def update(c, language='en'):
    '''Update the POT file and invoke the `sphinx-intl` `update` command

    Only used with `invoke intl.update`
    '''
    opts = "-b gettext"
    target = Path(c.sphinx.target).parent / 'gettext'
    if language == 'en':
        _clean(c)
        build(c, target=target, opts=opts)
    else:
        if not Path(target).exists():
            build(c, target=target, opts=opts)
        c.run(
            f'sphinx-intl update -p {target} -l {language}'
        )
        # for DIR in ['pages', 'posts', 'shop']:
        #     rmtree(f'locales/{language}/LC_MESSAGES/{DIR}/')


def _site(name, help_part):
    self = sys.modules[__name__]
    if name == 'docs':
        out = ''
    else:
        out = name
    coll = Collection.from_module(
        self,
        name=name,
        config={"sphinx": {
            "source": name,
            "target": f"output/html/{out}"
        }},
    )
    coll.__doc__ = f"Tasks for building {help_part}"
    coll["build"].__doc__ = f"Build {help_part}"
    return coll


# Usage doc/API site (published as e.g. docs.myproject.org)
docs = _site("docs", "the main site.")
write = _site("write", "the main site.")

ns = Collection(docs, write)



import re
from dataclasses import dataclass
from string import Template as _Template


@dataclass
class Tempalte:
    '''HTML 全局属性'''
    is_null: bool = False
    template: str = \
        '''<${tag} 
    accesskey="${accesskey}"
    autocapitalize="${autocapitalize}"
    autofocus="${autofocus}"
    class="${className}"
    contenteditable="${contenteditable}"
    dir="${dirName}"
    draggable="${draggable}"
    enterkeyhint="${enterkeyhint}"
    exportparts="${exportparts}"
    hidden="${hidden}"
    id="${id}"
    inputmode="${inputmode}"
    is="${isAction}"
    itemid="${itemid}"
    itemprop="${itemprop}"
    itemref="${itemref}"
    itemscope="${itemscope}"
    itemtype="${itemtype}"
    lang="${lang}"
    nonce="${nonce}"
    part="${part}"
    slot="${slot}"
    spellcheck="${spellcheck}"
    style="${style}"
    tabindex="${tabindex}"
    title="${title}"
    translate="${translate}" ${params}>
    ${content}</${tag}>'''.replace('\n', '').strip()

    def __post_init__(self):
        if self.is_null:
            template = self.template.replace('${content}</${tag}>', '')
        else:
            template = self.template
        self.template = re.sub(r"\s+", " ", template)

    def substitute(self, config):
        return _Template(self.template).substitute(config)


from tkinter import ttk
from tkinter import Toplevel


class TipWindow(Toplevel):
    def __init__(self, master, **kw):
        """创建带有工具提示文本的 topoltip 窗口"""
        super().__init__(master, **kw)
        self._custom(master)

    def _custom(self, widget):
        '''定制窗口属性
        
        Args:
            widget: tkinter 小部件 或者 tkinter.ttk 小部件
        '''
        ## 隐藏窗体的标题、状态栏等
        self.overrideredirect(True)
        ## 保持在主窗口的上面
        self.attributes("-toolwindow", 1)  # 也可以使用 `-topmost`
        self.attributes("-alpha", 0.92857142857)    # 设置透明度为 13/14

    def _label_params(self, text, textvariable):
        '''创建用来显示的标签'''
        return {
            'textvariable': textvariable,
            'text': text,
            'justify': 'left',
            'background': 'lightyellow',
            'relief': 'solid',
            'borderwidth': 1
        }


class ToolTip:
    '''针对指定的 widget 创建 tooltip

    参考：https://stackoverflow.com/a/36221216 
        以及 https://pysimplegui.readthedocs.io/en/latest/
    '''

    def __init__(self, widget, text=None, textvariable=None, timeout=300, offset=(0, -20), **kw):
        # 设置 用户参数
        self.widget = widget
        self.text = text
        self.textvariable = textvariable
        self.timeout = timeout
        self.offset = offset
        self._init_params()
        # 绑定事件
        self.widget.bind("<Enter>", self.enter)
        self.widget.bind("<Leave>", self.leave)
        self.widget.bind("<ButtonPress>", self.leave)

    def _init_params(self):
        '''内部参数的初始化'''
        self.id_after = None
        self.x, self.y = 0, 0
        self.tip_window = None

    def update_cursor(self, event):
        '''设定 鼠标光标的位置坐标 (x,y)'''
        self.x = event.x
        self.y = event.y

    def unschedule(self):
        '''取消用于鼠标悬停时间的计时器'''
        if self.id_after:
            self.widget.after_cancel(self.id_after)
        self.id_after = None

    def wm_geometry(self):
        '''转换为电脑桌面的坐标'''
        new_x = self.widget.winfo_rootx() + self.x + self.offset[0]
        new_y = self.widget.winfo_rooty() + self.y + self.offset[1]
        return new_x, new_y

    def show_tip(self):
        """创建带有工具提示文本的 topoltip 窗口"""
        if self.tip_window:
            return
        self.tip_window = TipWindow(self.widget)
        new_x, new_y = self.wm_geometry()
        self.tip_window.wm_geometry("+%d+%d" % (new_x, new_y))
        params = self.tip_window._label_params(
            self.text, self.textvariable)
        tip_label = ttk.Label(self.tip_window, **params)
        tip_label.grid(sticky='nsew')

    def schedule(self):
        """
        安排计时器以计时鼠标悬停的时间
        """
        self.id_after = self.widget.after(self.timeout, self.show_tip)

    def enter(self, event):
        """
        鼠标进入 widget 的回调函数
        
        Args:
            event:  来自于 tkinter，有鼠标的 x,y 坐标属性
        """
        self.unschedule()
        self.update_cursor(event)
        self.schedule()

    def hide_tip(self):
        """销毁 tooltip window"""
        if self.tip_window:
            self.tip_window.destroy()
        self.tip_window = None

    def leave(self, event):
        """
        鼠标离开 widget 的销毁 tooltip window
         
        Args:
            event:  来自于 tkinter，没有被使用
        """
        self.unschedule()
        self.hide_tip()



'''Some of the actions related to the graph. see https://www.jianshu.com/p/011a318ffd1e'''
from tkinter import Canvas as _Canvas, StringVar, ttk


class Canvas(_Canvas):
    '''Graphic elements are composed of line(segment), rectangle, ellipse, and arc.'''

    def __init__(self, master=None, cnf={}, **kw):
        '''The base class of all graphics frames.

        Args:
            master: a widget of tkinter or tkinter.ttk.
        '''
        super().__init__(master, cnf, **kw)

    def draw_graph(self, graph_type, direction, color='blue', width=1, tags=None, **kwargs):
        '''Draw basic graphic elements.

        Args:
            direction: Specifies the orientation of the graphic element. 
                Union[int, float] -> (x_0,y_0,x_,y_1), (x_0, y_0) refers to the starting point of 
                the reference brush (i.e., the left mouse button is pressed), and (x_1, y_1) refers to 
                the end position of the reference brush (i.e., release the left mouse button).
                Multipoint sequences are supported for 'line' and 'polygon',
                for example ((x_0, y_0), (x_1, y_1), (x_2, y_2)).

            graph_type: Types of graphic elements.
                (str) 'rectangle', 'oval', 'line', 'arc'(That is, segment), 'polygon'.
                Note that 'line' can no longer pass in the parameter 'fill', and 
                the remaining graph_type cannot pass in the parameter 'outline'.

            color: The color of the graphic element.

            width: The width of the graphic element.(That is, center fill)

            tags: The tags of the graphic element. 
                It cannot be a pure number (such as 1 or '1' or '1 2 3'), it can be a list, a tuple, 
                or a string separated by a space(is converted to String tupers separated by a blank space). 
                The collection or dictionary is converted to a string.
                
                Example:
                    ['line', 'graph'], ('test', 'g'), 'line',
                    ' line kind '(The blanks at both ends are automatically removed), and so on.

            style: Style of the arc in {'arc', 'chord', or 'pieslice'}.

        return: Unique identifier solely for graphic elements.
        '''
        if tags is None:
            if graph_type in ('rectangle', 'oval', 'line', 'arc'):
                tags = f"graph {graph_type}"
            else:
                tags = 'graph'

        com_kw = {'width': width, 'tags': tags}
        kw = {**com_kw, 'outline': color}
        line_kw = {**com_kw, 'fill': color}
        if graph_type == 'line':
            kwargs.update(line_kw)
        else:
            kwargs.update(kw)
        func = eval(f"self.create_{graph_type}")
        graph_id = func(*direction, **kwargs)
        return graph_id

class Drawing(Canvas):
    '''Create graphic elements (graph) including rectangular boxes (which can be square points), ovals (circular points), and segments.
        
    Press the left mouse button to start painting, release the left mouse button for the end of the painting.
    '''

    def __init__(self, master=None, selector=None, cnf={}, **kw):
        '''Click the left mouse button to start painting, release the left mouse button to complete the painting.

        Args:
            selector: The graphics selector, which is an instance of Selector.
        '''
        super().__init__(master, cnf, **kw)
        self.selector = selector

        self._init_params()
        self._draw_bind()

    def _draw_bind(self):
        self.bind("<1>", self.update_xy)
        self.bind("<ButtonRelease-1>", self.draw)

    def _init_params(self):
        self.x = self.y = 0

    def update_xy(self, event):
        '''Press the left mouse button to record the coordinates of the left mouse button'''
        self.x = self.canvasx(event.x)
        self.y = self.canvasy(event.y)

    def get_bbox(self, event):
        x0, y0 = self.x, self.y  # The upper-left coordinates of the graph
        self.update_xy(event)
        x1, y1 = self.x, self.y  # Lower-right coordinates of the graph
        bbox = x0, y0, x1, y1
        return bbox

    def draw(self, event):
        '''Release the left mouse button to finish painting.'''
        self.configure(cursor="arrow")
        bbox = self.get_bbox(event)
        self.create_graph(bbox)

    def create_graph(self, bbox):
        '''Create a graphic.

        Args:
            bbox: (x0,y0,x1,y1)
        '''
        x0, y0, x1, y1 = bbox
        cond1 = x0 == x1 and y0 == y1 and 'point' not in self.selector.graph_type
        cond2 = 'point' in self.selector.graph_type and (x0 != x1 or y0 != y1)
        if cond1 or cond2:
            return
        else:
            self.draw_graph(self.selector.graph_type.split('_')
                            [0], bbox, self.selector.color)

class TrajectoryDrawing(Drawing):
    '''Draw based on the mouse's trajectory.

    Click the left mouse button to start painting, move the 
        left mouse button 'after_time' after the completion of painting.
    '''

    def __init__(self, master,  selector, after_time=1000, cnf={}, **kw):
        super().__init__(master, selector, cnf, **kw)
        self.after_time = after_time
        self.bind("<1>", self.update_xy)
        self.bind("<ButtonRelease-1>", self.update_xy)
        self.bind("<Button1-Motion>", self.draw)

    def draw(self, event):
        '''Release the left mouse button to finish painting.'''
        self.configure(cursor="arrow")
        self.after(self.after_time)
        bbox = self.get_bbox(event)
        self.create_graph(bbox)

class Graph(Drawing):
    '''Graph is a drawing tool for manipulating graphic objects.'''
    def __init__(self, master,  selector, cnf={}, **kw):
        super().__init__(master, selector, cnf, **kw)
        self.selected_tags = None
        self.edit_var = StringVar()

    @property
    def tags_dict(self):
        return {
            'Single element': 'current',
            'All elements': 'all',
            'All graphic elements': 'graph',
            '⬜': 'rectangle',
            '⚪': 'oval',
            '⸺': 'line',
            '⯀': 'rectangle_point',
            '●': 'oval_point'
        }

    def move_strides(self, event):
        x, y = self.x, self.y
        self.update_xy(event)
        x_move = self.x - x
        y_move = self.y - y
        return x_move, y_move

    def move_graph(self, event):
        x_move, y_move = self.move_strides(event)
        self.move(self.selected_tags, x_move, y_move)

    def delete_graph(self, event):
        self.delete(self.selected_tags)

    def select_graph(self, event, edit_option):
        self.configure(cursor="target")
        self.update_xy(event)
        tags = self.tags_dict[edit_option]
        if tags == 'current':
            self.selected_tags = self.find_withtag(tags)
        else:
            self.selected_tags = tags

    def bind_graph(self):
        self.unbind('<ButtonRelease-1>')
        self.unbind('<1>')
        edit = self.edit_var.get()
        if edit == 'drawing':
            self._draw_bind()  # reset bind
        elif 'move' in edit:
            self.bind('<1>', lambda event: self.select_graph(
                event, edit.split('/')[1]))
            self.bind('<ButtonRelease-1>', self.move_graph)
        elif 'delete' in edit:
            self.bind('<1>', lambda event: self.select_graph(
                event, edit.split('/')[1]))
            self.bind('<ButtonRelease-1>', self.delete_graph)

class GraphScrollable(Graph):
    '''Pin the picture and label it on it.'''

    def __init__(self, master,  selector, cnf={}, **kw):
        super().__init__(master, selector, cnf, **kw)
        self.coord_var = StringVar()
        self.coord_label = ttk.Label(self, textvariable=self.coord_var)
        self._set_scroll()
        self._scroll_command()
        self.configure(xscrollcommand=self.scroll_x.set,
                       yscrollcommand=self.scroll_y.set)
        self.bind("<3>", self.mouse_motion)
        self.bind("<Configure>", self.resize)
        self.update_idletasks()
        #self.minsize(self.winfo_width(), self.winfo_height())
        self.scroll_x.pack(side='top', fill='x')
        self.pack(side='left', expand='yes', fill='both')
        self.scroll_y.pack(side='right', fill='y')
        self.coord_label.pack(side='bottom', fill='x')

    def draw(self, event):
        '''Release the left mouse button to finish painting.'''
        self.configure(cursor="arrow")
        bbox = self.get_bbox(event)
        self.create_graph(bbox)
        self.coord_var.set(f'coordinate: {bbox[:2]} to {bbox[2:]}')

    def select_graph(self, event, edit_option):
        self.configure(cursor="target")
        self.update_xy(event)
        tags = self.tags_dict[edit_option]
        image_tags = self.find_withtag('image')
        selected_tags = self.find_withtag(tags)
        self.selected_tags = set(selected_tags) - set(image_tags)

    def move_graph(self, event):
        x_move, y_move = self.move_strides(event)
        for tag in self.selected_tags:
            self.move(tag, x_move, y_move)

    def delete_graph(self, event):
        for tag in self.selected_tags:
            self.delete(tag)

    def _set_scroll(self):
        self.scroll_x = ttk.Scrollbar(self, orient='horizontal')
        self.scroll_y = ttk.Scrollbar(self, orient='vertical')

    def mouse_motion(self, event):
        x = event.x
        y = event.y
        text = f"coordinate: ({x}, {y})|| {self.canvasx(x), self.canvasy(y)}"
        self.coord_var.set(text)

    def _scroll_command(self):
        self.scroll_x['command'] = self.xview
        self.scroll_y['command'] = self.yview

    def resize(self, event):
        region = self.bbox('all')
        self.configure(scrollregion=region)




'''Use help reference at https://www.jianshu.com/p/c6a2b400d0b9

Selector and SelectorFrame are developer-oriented and do not 
    require users to pay attention to their implementation details.
'''
from tkinter import ttk, StringVar

from .base import Canvas


class Selector(Canvas):
    '''A selection icon that sets the shape and color of the graphic.

    Example:
    ======================
    from tkinter import Tk
    root =Tk()
    select = Selector(root)
    select.grid()
    root.mainloop()
    '''
    colors = 'red', 'blue', 'black', 'purple', 'green', 'skyblue', 'yellow', 'white', 'orange', 'pink'
    shapes = 'rectangle', 'oval', 'line', 'oval_point', 'rectangle_point'

    def __init__(self, master=None, cnf={}, **kw):
        '''The base class of all graphics frames.

        :param master: a widget of tkinter or tkinter.ttk.
        '''
        super().__init__(master, cnf, **kw)
        self.start, self.end = 15, 50
        self.create_color()
        self.create_shape()
        self.dtag('graph')

    def create_color(self):
        '''Set the color selector'''
        self.create_text((self.start, self.start),
                         text='color', font='Times 15', anchor='w')
        self.start += 10
        for k, color in enumerate(Selector.colors):
            t = 7+30*(k+1)
            direction = self.start+t, self.start-20, self.end+t, self.end-20
            tags = f"color {color}"
            self.draw_graph('rectangle', direction,
                            'yellow', tags=tags, fill=color)
        self.dtag('rectangle')

    def create_shape(self):
        '''Set the shape selector'''
        self.create_text((self.start-10, self.start+30),
                         text='shape', font='Times 15', anchor='w')
        for k, shape in enumerate(Selector.shapes):
            t = 7+30*(k+1)
            direction = self.start+t, self.start+20, self.end+t, self.end+20
            fill = 'blue' if 'point' in shape else 'white'
            width = 10 if shape == 'line' else 1
            kw = {
                'width': width,
                'fill': fill,
                'tags': f"shape {shape}"
            }
            self.draw_graph(shape.split('_')[0], direction, 'blue', **kw)


class SelectorFrame(ttk.Frame):
    '''Binding the left mouse button function of the graphics selector to achieve the color and 
        shape of the graphics change.

    Example:
    ===============================================
    from tkinter import Tk
    root = Tk()
    selector = SelectorFrame(root)
    selector.layout()
    selector.grid()
    root.mainloop()
    '''

    def __init__(self, master=None, graph_type='rectangle', color='blue', **kw):
        '''The base class of all graphics frames.

        :param master: a widget of tkinter or tkinter.ttk.
        :param graph_type: The initial shape value of the graph.
        :param color: The initial color value of the graph.
        '''
        super().__init__(master, **kw)
        self.color = color
        self.graph_type = graph_type
        self.selector = Selector(
            self, background='lightgreen', width=360, height=80)
        [self.color_bind(self.selector, color) for color in Selector.colors]
        [self.graph_type_bind(self.selector, graph_type)
         for graph_type in Selector.shapes]
        self.info_var = StringVar()
        self.info = ttk.Label(self, textvariable=self.info_var)
        # 布局
        self.selector.grid(row=0, column=0)
        self.info.grid(row=1, column=0)

    def update_info(self):
        '''Update info information.'''
        if self.color or self.graph_type:
            text = f"You Selected: {self.color},{self.graph_type}"
            self.info_var.set(text)

    def update_color(self, new_color):
        '''Update color information.'''
        self.color = new_color
        self.update_info()

    def update_graph_type(self, new_graph_type):
        '''Update graph_type information.'''
        self.graph_type = new_graph_type
        self.update_info()

    def color_bind(self, canvas, color):
        canvas.tag_bind(color, '<1>', lambda e: self.update_color(color))

    def graph_type_bind(self, canvas, graph_type):
        canvas.tag_bind(graph_type, '<1>',
                        lambda e: self.update_graph_type(graph_type))



def lfs_url(repo, filename, branch='main'):
    '''
    Example
    ===========
    repo: 'SanstyleLab/plotly-datasets'
    branch: 'main'
    filename: 'gapminderDataFiveYear.csv'
    '''
    root = 'https://media.githubusercontent.com/media'
    url = f'{root}/{repo}/{branch}/{filename}'
    return url


class _EventTarget:
    '''https://developer.mozilla.org/en-US/docs/Web/API/EventTarget'''
    NotImplemented


class _Node(_EventTarget):
    '''https://developer.mozilla.org/en-US/docs/Web/API/Node'''
    NotImplemented


class _Element(_Node):
    '''ref of https://developer.mozilla.org/en-US/docs/Web/API/Element'''
    NotImplemented


#from functools import lru_cache
from .base import TrajectoryDrawing

class Rectangle:
    def __init__(self, bbox):
        self.x0, self.y0, self.x1, self.y1 = bbox
        self.bunch = {
            'left_top_corner': (self.x0, self.y0),
            'top_middle': (self.center[0], self.y0),
            'right_top_corner': (self.x1, self.y0),
            'right_middle': (self.x1, self.center[1]),
            'right_bottom_corner': (self.x1, self.y1),
            'bottom_middle': (self.center[0], self.y1),
            'left_bottom_corner': (self.x0, self.y1),
            'left_middle': (self.x0, self.center[1])
        }

    @property
    def grad_x(self):
        return self.x1 - self.x0

    @property
    def grad_y(self):
        return self.y1 - self.y0

    @property
    def width(self):
        return abs(self.grad_x)

    @property
    def height(self):
        return abs(self.grad_y)

    @property
    def center(self):
        x = (self.x0 + self.x1)/2
        y = (self.y0 + self.y1)/2
        return x, y

    def __contains__(self, point):
        x, y = point
        x_cond = x in range(self.x0, self.x1)
        y_cond = y in range(self.y0, self.y1)
        return x_cond and y_cond

    def __lt__(self, other):
        '''self < other'''
        return self.width < other.width or self.height < other.height

    def __le__(self, other):
        '''self < other'''
        return self.width <= other.width or self.height <= other.height

class DrawRectangle(TrajectoryDrawing):
    def __init__(self, master=None, alpha=0.2, color='blue', graph_type='rectangle',
                 min_size=7, mask_color='red', cnf={}, **kw):
        '''
        '''
        super().__init__(master, graph_type, color, min_size, cnf, **kw)
        self.alpha= alpha
        self.mask_color = mask_color
        self.rectangle_selector = []
        self.tag_bind('mask', "<ButtonPress-1>", self.scroll_start)
        self.tag_bind('mask', "<B1-Motion>", self.scroll_move)

    def scroll_start(self, event):
        self.scan_mark(event.x, event.y)

    def scroll_move(self, event):
        # self.move('current', w, h)
        # graph_id = self.find_withtag('current')
        # self.select_from(graph_id, 1)
        self.scan_dragto(event.x, event.y, gain=1)

    def create_mask(self, size, alpha, fill):
        '''设置透明蒙版'''
        fill = self.master.winfo_rgb(fill) + (alpha,)
        return Image.new('RGBA', size, fill)
        
    def draw_mask(self, x1, y1, x2, y2):
        size = x2-x1, y2-y1
        alpha = int(self.alpha * 255)
        image = self.create_mask(size, alpha, self.mask_color)
        self.rectangle_selector.append(ImageTk.PhotoImage(image))
        return self.create_image(x1, y1, image=self.rectangle_selector[-1], anchor='nw', tags='mask')

    def finish_drawing(self, *event,  **kw):
        if self.graph_type and self.record_bbox:
            if graph_id := self.drawing(width=1, tags=None, **kw):
                bbox = self.bbox(graph_id)
                mask = self.draw_mask(*bbox)
        self.reset(event)



import logging
from logging.handlers import RotatingFileHandler

def config_logging(
    filename: str,
    logger_name: str="logger",
    filter_mod_names: set=set(),
    default_filter_mod_names: set = {
        "matplotlib", "PIL", "asyncio", 
        "urllib3", "markdown_it",
        "ablog"
    },
    file_formatter: str="%(levelname)s|%(asctime)s|%(name)s| >>> %(message)s\n|==>%(module)s.%(funcName)s@: %(pathname)s:%(lineno)d\n\n",
    stream_formatter: str="%(levelname)s|%(asctime)s|%(name)s| >>> %(message)s",
    **kwargs):
    """配置 logging

    Args:
        filename: 日志保存路径
        logger_name: 日志名称
        filemode: 写入到文件的日志写入模式
        filter_mod_names: 自定义过滤日志 debug 模式对应的模块名称列表
        default_filter_mod_names: 默认过滤日志 debug 模式对应的模块名称列表
        file_formatter: 日志文件配置
        stream_formatter: 控制台打印配置
        rich_kwargs: 富文本模式，可为空
            即开启如下模式
            ```
            >>> from rich.logging import RichHandler
            >>> ch = RichHandler(**rich_kwargs)
            ```
    """
    logger = logging.getLogger(logger_name)
    # 禁用一些 debug 信息
    for mod_name in filter_mod_names|default_filter_mod_names:
        _logger = logging.getLogger(mod_name)
        _logger.setLevel(logging.WARNING)
    logger.setLevel(logging.DEBUG)
    fh = RotatingFileHandler(filename, **kwargs)
    fh.setFormatter(logging.Formatter(file_formatter))
    # fh.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO) # 或者 logging.ERROR
    ch_formatter = logging.Formatter(stream_formatter)
    ch.setFormatter(ch_formatter)
    logging.getLogger("").addHandler(ch)
    


from __future__ import annotations
from dataclasses import dataclass
from contextlib import ContextDecorator
import logging
import time
import numpy as np

@dataclass
class TimerContext(ContextDecorator):
    """追踪上下文信息

    Args:
        name: 计时器名称
    """
    name: str = "Timer"

    def __post_init__(self):
        # 记录开始时间
        self._times = []  # 记录时间戳
        self.times = []  # 记录时间段

    def __enter__(self):
        logging.debug(f"Entering {self.name}.")
        self._times.append(time.time())
        return self

    @property
    def runtime(self):
        return (self._times[-1] - self._times[-2]) * 1000

    def __exit__(self, *exc_details):
        self._times.append(time.time())
        self.times.append(self.runtime)
        logging.debug(f"Run time: {self.times[-1]:.7g} ms.")
        logging.debug(f"Exiting {self.name}.")
        return False

    def reset(self):
        """重置记录"""
        self._times = []
        self.times = []

    def __len__(self):
        """返回运行次数"""
        return len(self.times)

    def sum(self):
        """返回时间总和"""
        return sum(self.times)

    def avg(self):
        """返回平均时间"""
        return self.sum() / len(self)

    def cumsum(self):
        """返回累计时间"""
        return np.array(self.times).cumsum().tolist()



from functools import update_wrapper
from bytecode import Bytecode, Instr
import logging
# logging.basicConfig(level=logging.INFO)


class LoggedAccess:
    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = '_' + name

    def __get__(self, obj, objtype=None):
        value = getattr(obj, self.private_name)
        logging.debug('Accessing %r giving %r', self.public_name, value)
        return value

    def __set__(self, obj, value):
        logging.debug('Updating %r to %r', self.public_name, value)
        setattr(obj, self.private_name, value)


class trace:
    cache = {}
    is_activate = False
    variable = LoggedAccess()
    def __init__(self, variable="result"):
        self.variable = variable

    def __call__(self, func):
        if not type(self).is_activate:
            return func
        func = self.transform(func)
        def wrapper(*args, **kwargs):
            res, values = func(*args, **kwargs)
            type(self).cache[func.__qualname__].append((self.variable, values))
            return res
        return update_wrapper(wrapper, func)

    def transform(self, func):
        type(self).cache[func.__qualname__] = []
        c = Bytecode.from_code(func.__code__)
        extra_code = [
            Instr('STORE_FAST', '_res'),
            Instr('LOAD_FAST', self.variable),
            Instr('STORE_FAST', '_value'),
            Instr('LOAD_FAST', '_res'),
            Instr('LOAD_FAST', '_value'),
            Instr('BUILD_TUPLE', 2),
            Instr('STORE_FAST', '_result_tuple'),
            Instr('LOAD_FAST', '_result_tuple'),
         ]
        c[-1:-1] = extra_code
        func.__code__ = c.to_code()
        return func

    @classmethod
    def clear(cls):
        cls.cache.clear()

    @classmethod
    def activate(cls):
        cls.is_activate = True



from string import Template
from urllib.parse import urlencode

element_template = Template(
    '''<${tag} class="${className}" type="${typeName}" ${config}">${content}</${tag}>''')
frame_template = Template(
    '''<${tag} class="${className}" type="${typeName}" src="${src}${params}" ${config}>''')


class Embed:
    """
    Generic class to embed an embed in an IPython notebook
    """
    _embed = """<embed type="{typeName}" class="{className}" width="{width}" height="{height}" src="{src}{params}">"""

    def __init__(self, src, height, width='100%', className='w3-card w3-pale-blue', typeName='text/html', **kwargs):
        self.typeName = typeName
        self.className = className
        self.src = src
        self.width = width
        self.height = height
        self.params = kwargs

    def _repr_html_(self):
        """return the embed iframe"""
        if self.params:
            params = "?" + urlencode(self.params)
        else:
            params = ""
        config = {
            'typeName': self.typeName,
            'src': self.src,
            'className': self.className,
            'width': self.width,
            'height': self.height,
            'params': params
        }
        return self._embed.format(**config)